#    Copyright 2020 Connor Ferster

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import List, Any


def sympy_cell_line_lists(cell: str) -> List[List[str]]:
    """
    Converts sympy cell to list of lists of str
    """
    raw_lines = cell.split("\n")
    raw_line_components = [
        [elem.strip() for elem in line.split("=")] for line in raw_lines
    ]
    return raw_line_components


def test_sympy_parents(sympy_cls: str, parents: tuple) -> bool:
    """
    Returns True if 'sympy_cls' is in 'bases'.
    False, otherwise.
    """
    return any([sympy_cls in str(parent) for parent in parents])


def test_for_sympy_symbol(obj_str: str, var_dict: dict) -> bool:
    """
    Return True if 'obj_str' is in 'var_dict' and 'obj_str' represents
    a sympy object.
    """
    if obj_str not in var_dict:
        False
    else:
        obj = var_dict[obj_str]
        obj_type = type(obj)
        parents = obj_type.__mro__
        sympy_obj_test = test_sympy_parents("sympy.core.basic.Basic", parents)
        sympy_sym_test = test_sympy_parents("sympy.core.symbol.Symbol", parents)
        return sympy_obj_test and sympy_sym_test


def test_for_sympy_expr(obj_str: str, var_dict: dict) -> bool:
    """
    Return True if 'obj_str' is in 'var_dict' and 'obj_str' represents
    a sympy object.
    """
    if obj_str not in var_dict:
        False
    else:
        obj = var_dict[obj_str]
        obj_type = type(obj)
        parents = obj_type.__mro__
        return test_sympy_parents("sympy.core.basic.Basic", parents)


def test_for_sympy_eqn(obj_str: str, var_dict: dict) -> bool:
    """
    Return True if 'obj_str' is in 'var_dict' and 'obj_str' represents
    a sympy object.
    """
    if obj_str not in var_dict:
        False
    else:
        obj = var_dict[obj_str]
        obj_type = type(obj)
        parents = obj_type.__mro__
        return test_sympy_parents("sympy.core.relational.Equality", parents)


def convert_sympy_obj_to_py_str(obj_str: str, var_dict: dict) -> str:
    """
    Returns the sympy obj represented by the dict key 'obj_str', retrieved from
    'var_dict' as a python code string.
    """
    return str(var_dict["obj_str"])


def get_sympy_obj(obj_str: str, var_dict: dict) -> Any:
    """
    Returns the object represented by 'obj_str' from 'var_dict'
    """
    sp_obj = var_dict[obj_str]
    if isinstance(sp_obj, list) and len(sp_obj) == 1:
        return sp_obj[0]
    return sp_obj


def convert_sympy_cell_to_py_cell(cell: str, var_dict: dict) -> str:
    """
    Returns 'cell' converted from a multiline string representing a bunch 
    of sympy expressions and equality objects to a multiline string of 
    equivalent, representative python code for rendering by handcalcs.
    """
    acc = []
    lines = cell.split("\n")
    for line in lines:
        # try:
        if "=" in line:
            lhs, rhs = line.split("=", 1)
            obj_str = rhs.strip()
            if test_for_sympy_eqn(obj_str, var_dict):
                sym_obj = get_sympy_obj(obj_str, var_dict)
                lhs = sym_obj.lhs
                rhs = sym_obj.rhs
                acc.append(str(lhs) + "=" + str(rhs))
            elif test_for_sympy_expr(obj_str, var_dict):
                sym_obj = get_sympy_obj(obj_str, var_dict)
                acc.append(lhs + "=" + str(sym_obj))
            else:
                acc.append(line)
        else:
            obj_str = line.strip()
            if test_for_sympy_eqn(obj_str, var_dict):
                sym_obj = get_sympy_obj(obj_str, var_dict)
                lhs = sym_obj.lhs
                rhs = sym_obj.rhs
                acc.append(str(lhs) + "=" + str(rhs))
            elif test_for_sympy_symbol(obj_str, var_dict):
                sym_obj = get_sympy_obj(obj_str, var_dict)
                acc.append(str(sym_obj))
            elif test_for_sympy_expr(obj_str, var_dict):
                raise ValueError(
                    f"The result of a sympy expr must be assigned to a new variable, e.g. x = {line}"
                )
            else:
                acc.append(line)
        # except:
        #    raise ValueError(f"%%render sympy: Should only be used for a cell filled with sympy objects, not: {line}")
    return "\n".join(acc)



__all__ = ["handcalc"]

from functools import wraps
import inspect
import innerscope
from handcalcs.handcalcs import LatexRenderer


def handcalc(
    override: str = "", precision: int = 3, left: str = "", right: str = "", decimal_separator:str = ".", jupyter_display: bool = False
):
    # @wraps(func)
    def handcalc_decorator(func):

        def wrapper(*args, **kwargs):
            line_args = {"override": override, "precision": precision}
            func_source = inspect.getsource(func)
            cell_source = _func_source_to_cell(func_source)
            # use innerscope to get the values of locals, closures, and globals when calling func
            scope = innerscope.call(func, *args, **kwargs)
            LatexRenderer.dec_sep = decimal_separator
            renderer = LatexRenderer(cell_source, scope, line_args)
            latex_code = renderer.render()
            if jupyter_display:
                try:
                    from IPython.display import Latex, display
                except ModuleNotFoundError:
                    ModuleNotFoundError(
                        "jupyter_display option requires IPython.display to be installed."
                    )
                display(Latex(latex_code))
                return scope.return_value

            # https://stackoverflow.com/questions/9943504/right-to-left-string-replace-in-python
            latex_code = "".join(latex_code.replace("\\[", "", 1).rsplit("\\]", 1))
            return (left + latex_code + right, scope.return_value)

        return wrapper

    return handcalc_decorator


def _func_source_to_cell(source: str):
    """
    Returns a string that represents `source` but with no signature, doc string,
    or return statement.
    
    `source` is a string representing a function's complete source code.
    """
    source_lines = source.split("\n")
    acc = []
    doc_string = False
    for line in source_lines:
        if (not doc_string
            and line.lstrip(' \t').startswith('"""')
            and line.lstrip(' \t').rstrip().endswith('"""',3)):
            doc_string = False
            continue
        elif (not doc_string
            and line.lstrip(' \t').startswith('"""')
            and not line.lstrip(' \t').rstrip().endswith('"""',3)):
            doc_string = True
            continue
        elif doc_string and '"""' in line:
            doc_string = False
            continue
        if (
            "def" not in line
            and not doc_string
            and "return" not in line
            and "@" not in line
        ):
            acc.append(line)
    return "\n".join(acc)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['handcalcs']

package_data = \
{'': ['*'], 'handcalcs': ['templates/html/*', 'templates/latex/*']}

install_requires = \
['innerscope>=0.2.0,<0.3.0',
 'more-itertools>=8.5.0,<9.0.0',
 'nbconvert>=5.6.1,<6.0.0',
 'pyparsing>=2.4.7,<3.0.0']

setup_kwargs = {
    'name': 'handcalcs',
    'version': '1.2.1',
    'description': 'Python calculations in Jupyter as though you wrote them by hand.',
    'long_description': None,
    'author': 'Connor Ferster',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/connorferster/handcalcs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#    Copyright 2020 Connor Ferster

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import sys
from . import handcalcs as hand
from . import sympy_kit as s_kit

try:
    from IPython.core.magic import (
        Magics,
        magics_class,
        cell_magic,
        register_cell_magic,
        register_line_magic,
    )
    from IPython import get_ipython
    from IPython.display import Latex, Markdown, display
    from IPython.utils.capture import capture_output
except ImportError:
    pass


try:
    ip = get_ipython()
    cell_capture = capture_output(stdout=True, stderr=True, display=True)
except AttributeError:
    raise ImportError(
        "handcalcs.render is intended for a Jupyter environment."
        " Use 'from handcalcs import handcalc' for the decorator interface."
    )


def parse_line_args(line: str) -> dict:
    """
    Returns a dict that represents the validated arguments
    passed in as a line on the %%render or %%tex cell magics.
    """
    valid_args = ["params", "long", "short", "sympy", "symbolic", "_testing"]
    # valid_args = ["params", "long", "short", "sympy", "symbolic", "_testing"]
    sympy_arg = ["sympy"]
    line_parts = line.split()
    parsed_args = {"override": "", "precision": "", "sympy": False}
    # parsed_args = {
    #     "override": "",
    #     "precision": "",
    # }
    precision = ""
    for arg in line_parts:
        if arg.lower() in sympy_arg:
            parsed_args["sympy"] = True
            continue
        for valid_arg in valid_args:
            if arg.lower() in valid_arg:
                parsed_args.update({"override": valid_arg})
                break
        try:
            precision = int(arg)
        except ValueError:
            pass
        if precision:
            parsed_args.update({"precision": precision})
    return parsed_args


@register_line_magic
def decimal_separator(line):
    if len(line) == 1:
        hand.LatexRenderer.dec_sep = line


@register_cell_magic
def render(line, cell):
    # Retrieve var dict from user namespace
    user_ns_prerun = ip.user_ns
    line_args = parse_line_args(line)

    if line_args["sympy"]:
        cell = s_kit.convert_sympy_cell_to_py_cell(cell, user_ns_prerun)

    # Run the cell
    with cell_capture:
        exec_result = ip.run_cell(cell)

    if not exec_result.success:
        return None

    # Retrieve updated variables (after .run_cell(cell))
    user_ns_postrun = ip.user_ns

    # Do the handcalc conversion
    renderer = hand.LatexRenderer(cell, user_ns_postrun, line_args)
    latex_code = renderer.render()

    # Display, but not as an "output"
    display(Latex(latex_code))

    if line_args["override"] == "_testing":
        return latex_code


@register_cell_magic
def tex(line, cell):
    # Retrieve var dict from user namespace
    user_ns_prerun = ip.user_ns
    line_args = parse_line_args(line)

    if line_args["override"] == "sympy":
        cell = s_kit.convert_sympy_cell_to_py_cell(cell, user_ns_prerun)

    # Run the cell
    with cell_capture:
        ip.run_cell(cell)

    # Retrieve updated variables (after .run_cell(cell))
    user_ns_postrun = ip.user_ns

    # Do the handcalc conversion
    renderer = hand.LatexRenderer(cell, user_ns_postrun, line_args)
    latex_code = renderer.render()

    # Print Latex Code
    print(latex_code)

    if line_args["override"] == "_testing":
        return latex_code


def load_ipython_extension(ipython):
    """This function is called when the extension is
    loaded. It accepts an IPython InteractiveShell
    instance. We can register the magic with the
    `register_magic_function` method of the shell
    instance."""
    ipython.register_magic_function(render, "cell")


# def unload_ipython_extension(ipython):
#     """This function is called when the extension is
#     loaded. It accepts an IPython InteractiveShell
#     instance. We can register the magic with the
#     `register_magic_function` method of the shell
#     instance."""
#     print(dir(ipython.magics_manager))
#     ipython.magics_manager.remove(render)
